=== modified file 'bzrlib/tests/tree_implementations/test_tree.py'
--- bzrlib/tests/tree_implementations/test_tree.py	2007-12-30 22:24:01 +0000
+++ bzrlib/tests/tree_implementations/test_tree.py	2007-12-31 05:35:39 +0000
@@ -130,6 +130,18 @@
                          set(['b-id', 'root-id', 'c-id', 'a-id']))
 
 
+class TestStoredKind(TestCaseWithTree):
+
+    def test_stored_kind(self):
+        tree = self.make_branch_and_tree('tree')
+        work_tree = self.make_branch_and_tree('wt')
+        tree = self.get_tree_no_parents_abc_content(work_tree)
+        tree.lock_read()
+        self.addCleanup(tree.unlock)
+        self.assertEqual('file', tree.stored_kind('a-id'))
+        self.assertEqual('directory', tree.stored_kind('b-id'))
+
+
 class TestFileContent(TestCaseWithTree):
 
     def test_get_file(self):

=== modified file 'bzrlib/tests/workingtree_implementations/test_workingtree.py'
--- bzrlib/tests/workingtree_implementations/test_workingtree.py	2007-12-30 22:24:01 +0000
+++ bzrlib/tests/workingtree_implementations/test_workingtree.py	2007-12-31 05:35:39 +0000
@@ -825,6 +825,17 @@
             expected_kind = names[i]
             self.assertEqual(expected_kind, actual_kind)
 
+    def test_stored_kind_with_missing(self):
+        tree = self.make_branch_and_tree('tree')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        self.build_tree(['tree/a', 'tree/b/'])
+        tree.add(['a', 'b'], ['a-id', 'b-id'])
+        os.unlink('tree/a')
+        os.rmdir('tree/b')
+        self.assertEqual('file', tree.stored_kind('a-id'))
+        self.assertEqual('directory', tree.stored_kind('b-id'))
+
     def test_missing_file_sha1(self):
         """If a file is missing, its sha1 should be reported as None."""
         tree = self.make_branch_and_tree('.')

=== modified file 'bzrlib/transform.py'
--- bzrlib/transform.py	2007-12-30 22:27:32 +0000
+++ bzrlib/transform.py	2007-12-31 05:35:39 +0000
@@ -880,7 +880,6 @@
             conflicts = self.find_conflicts()
             if len(conflicts) != 0:
                 raise MalformedTransform(conflicts=conflicts)
-        inv = self._tree.inventory
         inventory_delta = []
         child_pb = bzrlib.ui.ui_factory.nested_progress_bar()
         try:
@@ -892,8 +891,7 @@
                 child_pb.update('Apply phase', 0, 2)
                 self._apply_removals(inventory_delta, mover)
                 child_pb.update('Apply phase', 1, 2)
-                modified_paths = self._apply_insertions(inv, inventory_delta,
-                                                        mover)
+                modified_paths = self._apply_insertions(inventory_delta, mover)
             except:
                 mover.rollback()
                 raise
@@ -987,7 +985,7 @@
         finally:
             child_pb.finished()
 
-    def _apply_insertions(self, inv, inventory_delta, mover):
+    def _apply_insertions(self, inventory_delta, mover):
         """Perform tree operations that insert directory/inventory names.
         
         That is, create any files that need to be created, and restore from
@@ -1002,10 +1000,6 @@
             for num, (path, trans_id) in enumerate(new_paths):
                 new_entry = None
                 child_pb.update('adding file', num, len(new_paths))
-                try:
-                    kind = self._new_contents[trans_id]
-                except KeyError:
-                    kind = contents = None
                 if trans_id in self._new_contents or \
                     self.path_changed(trans_id):
                     full_path = self._tree.abspath(path)
@@ -1021,13 +1015,17 @@
                     if trans_id in self._new_contents:
                         modified_paths.append(full_path)
                         completed_new.append(trans_id)
-
-                if trans_id in self._new_id:
-                    if kind is None:
-                        kind = file_kind(self._tree.abspath(path))
+                file_id = self.final_file_id(trans_id)
+                if file_id is not None and (trans_id in self._new_id or
+                    trans_id in self._new_name or trans_id in self._new_parent
+                    or trans_id in self._new_executability):
+                    try:
+                        kind = self.final_kind(trans_id)
+                    except NoSuchFile:
+                        kind = self._tree.stored_kind(file_id)
                     if trans_id in self._new_reference_revision:
                         new_entry = inventory.TreeReference(
-                            self._new_id[trans_id],
+                            self.final_file_id(trans_id),
                             self._new_name[trans_id], 
                             self.final_file_id(self._new_parent[trans_id]),
                             None, self._new_reference_revision[trans_id])
@@ -1035,34 +1033,16 @@
                         new_entry = inventory.make_entry(kind,
                             self.final_name(trans_id),
                             self.final_file_id(self.final_parent(trans_id)),
-                            self._new_id[trans_id])
-                else:
-                    if trans_id in self._new_name or trans_id in\
-                        self._new_parent or\
-                        trans_id in self._new_executability:
-                        file_id = self.final_file_id(trans_id)
-                        if file_id is not None:
-                            entry = inv[file_id]
-                            new_entry = entry.copy()
-
-                    if trans_id in self._new_name or trans_id in\
-                        self._new_parent:
-                            if new_entry is not None:
-                                new_entry.name = self.final_name(trans_id)
-                                parent = self.final_parent(trans_id)
-                                parent_id = self.final_file_id(parent)
-                                new_entry.parent_id = parent_id
+                            self.final_file_id(trans_id))
+                    try:
+                        old_path = self._tree.id2path(new_entry.file_id)
+                    except errors.NoSuchId:
+                        old_path = None
+                    inventory_delta.append((old_path, path, new_entry.file_id,
+                                            new_entry))
 
                 if trans_id in self._new_executability:
                     self._set_executability(path, new_entry, trans_id)
-                if new_entry is not None:
-                    if new_entry.file_id in inv:
-                        old_path = inv.id2path(new_entry.file_id)
-                    else:
-                        old_path = None
-                    inventory_delta.append((old_path, path,
-                                            new_entry.file_id,
-                                            new_entry))
         finally:
             child_pb.finished()
         for trans_id in completed_new:

=== modified file 'bzrlib/tree.py'
--- bzrlib/tree.py	2007-12-30 22:24:01 +0000
+++ bzrlib/tree.py	2007-12-31 05:35:39 +0000
@@ -189,6 +189,14 @@
         raise NotImplementedError("Tree subclass %s must implement kind"
             % self.__class__.__name__)
 
+    def stored_kind(self, file_id):
+        """File kind stored for this file_id.
+
+        May not match kind on disk for working trees.  Always available,
+        even for missing files.
+        """
+        return self.kind(file_id)
+
     def path_content_summary(self, path):
         """Get a summary of the information about path.
         

=== modified file 'bzrlib/workingtree.py'
--- bzrlib/workingtree.py	2007-12-30 22:24:01 +0000
+++ bzrlib/workingtree.py	2007-12-31 05:35:39 +0000
@@ -1702,6 +1702,10 @@
     def kind(self, file_id):
         return file_kind(self.id2abspath(file_id))
 
+    def stored_kind(self, file_id):
+        """See Tree.stored_kind"""
+        return self.inventory[file_id].kind
+
     def _comparison_data(self, entry, path):
         abspath = self.abspath(path)
         try:

=== modified file 'bzrlib/workingtree_4.py'
--- bzrlib/workingtree_4.py	2007-12-30 22:24:01 +0000
+++ bzrlib/workingtree_4.py	2007-12-31 05:35:39 +0000
@@ -1609,6 +1609,10 @@
             raise errors.NoSuchId(tree=self, file_id=file_id)
         return dirstate.DirState._minikind_to_kind[entry[1][0]]
 
+    def stored_kind(self, file_id):
+        """See Tree.stored_kind"""
+        return self.kind(file_id)
+
     def path_content_summary(self, path):
         """See Tree.path_content_summary."""
         id = self.inventory.path2id(path)

